/* Silly program to "trivially" (ha!) store uid/gid numbers in correct place.
 * Version 2: Version 1 didn't work with LinuxPPC Release 5.
 * Version 3: Modified (formatting) + two small tweaks to get it to work on
 *            Solaris from version provided by
 *            Pascal DeMilly <pascal.contribs@newgenesys.com>
 *            "Needed a major cleanup as I saw too many cases where it
 *             will just not work".
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>

/* Make sure this matches with auto_uids.c */
typedef struct
{
  int mrkb;
  int auto_uida;
  int auto_uidd;
  int auto_uidl;
  int auto_uido;
  int auto_uidp;
  int auto_uidq;
  int auto_uidr;
  int auto_uids;
  int auto_gidq;
  int auto_gidn;
  int mrke;
} block_t;

int filebrand( int file, char *filename, block_t * );
int setuser( int *addr, char *username );
int setgroup( int *addr, char *groupname );
void usage();


int main( int argc, char **argv )
{
block_t master_block;
int n;
int fd;

/* Initialize users and groups */
master_block.mrkb = 0x6d726b62; /* Should be "mrkb" */
setuser( &master_block.auto_uida, "alias"  );
setuser( &master_block.auto_uidd, "qmaild" );
setuser( &master_block.auto_uidl, "qmaill" );
setuser( &master_block.auto_uido, "root"   );
setuser( &master_block.auto_uidp, "qmailp" );
setuser( &master_block.auto_uidq, "qmailq" );
setuser( &master_block.auto_uidr, "qmailr" );
setuser( &master_block.auto_uids, "qmails" );

setgroup( &master_block.auto_gidq, "qmail"   );
setgroup( &master_block.auto_gidn, "nofiles" );
master_block.mrke = 0x6d726b65; /* Should be "mrke" */

if ( argc < 2 )
   usage();

for ( n = 1; n < argc; n++ )
   {
   if ( (fd = open( argv[n], O_RDWR )) == -1 )
      {
      fprintf( stderr, "Can't open file %s\n", argv[n] );
      return 1;
      }
   /* fprintf( stderr, "Searching %s\n", argv[n] ); */
   if ( filebrand( fd, argv[n], &master_block ) )
      return 1;
   close( fd );
   }

return 0;
}

void usage( void )
{
fprintf( stderr, "Usage: brand file ...\n" );
exit( 1 );
}

off_t foundMarker( int file, off_t pos, block_t *blockPtr )
{
block_t b;
int n;

/* Seek to pos and read in the full block */

if ( lseek( file, pos, SEEK_SET ) != pos )
   {
   perror( "Error: foundMarker(seek1)" );
   return -1;
   }

/* Read a block */
if ( (n = read( file, &b, sizeof(b) ) != sizeof(b)) )
   {
   /* Not enough to hold a block so this isn't it.  Continue. */
   pos += n;
   }
 else
   {
   /* If the numbers don't match, something really wierd happened. */
   if ( blockPtr->mrkb != b.mrkb )
      {
      perror( "Error foundMarker: mrkb not at pos\n" );
      return -1;
      }

   if ( blockPtr->mrke == b.mrke )
      {
      if ( lseek( file, pos, SEEK_SET ) != pos )
         {
         perror( "Error foundMarker(seek2)" );
         return -1;
         }

      /* Now write out our updates */
      if ( write( file, blockPtr, sizeof(*blockPtr) ) != sizeof(*blockPtr) )
         {
         perror( "Error foundMarker(write)" );
         return -1;
         }

      pos += sizeof(*blockPtr);
      return -1; /* OK. We wrote it out. */
      }
   }

/* Keep from looping */
return pos + 1;
}

/* Search for beginning and end marker and brand integers in between */
int filebrand( int file, char *filename, block_t *blockPtr )
{
unsigned char buf [ sizeof(int) * 256 ]; // Buffer must be at least 1 int long.
int n;
int pos = 0;
block_t template_block;
int marker;

/* Go through each character of file, filling in buffer until string terminator
   found or buffer is full */
while ( (n = read( file, &buf[0], sizeof(buf) )) > sizeof(int) )
   {
   int i;

   /* Check buffer 1 char at a time looking for int that resembles mrkb.
      Stop before the end so we don't get a partial int.
   */
   for ( i = 0; i < (n - sizeof(int) + 1); i++, pos++ )
      {
      memcpy( &marker, &buf[i], sizeof(marker) );

      if ( marker == blockPtr->mrkb )
         {
         if ( (pos = foundMarker( file, pos, blockPtr )) < 0 )
            return 0;
          else
            break;
         }
      }

      /* Now reposition; we could have read a part of a int at the end of the
         buffer.  Also, if we had found the marker the position would have
         changed to the marker end + 1.
      */
      lseek( file, pos, SEEK_SET );
   }

/* We should have found SOMETHING! */
fprintf( stderr, "%s: Marker not found.\n", filename );

return 1;
}

setuser( int *addr, char *username )
{
struct passwd *myp;

if ( (myp = getpwnam( username )) == NULL )
   {
   fprintf( stderr, "Can't find user %s\n", username );
   exit( 1 );
   }

*addr = (int) (myp->pw_uid);
}

setgroup( int *addr, char *groupname )
{
struct group *myg;

if ( (myg = getgrnam( groupname )) == NULL )
   {
   fprintf( stderr, "Can't find group %s\n", groupname );
   exit( 1 );
   }

*addr = (int) (myg->gr_gid);
}
